@@ -1,4 +1,4 @@
-/* Copyright © 2007, 2008, 2009 Jakub Wilk
+/* Copyright © 2007, 2008, 2009, 2010 Jakub Wilk
  * Copyright © 2009 Mateusz Turcza
  *
  * This package is free software; you can redistribute it and/or modify
@@ -174,6 +174,7 @@ Config::Config()
   this->bg_slices = NULL;
   this->file_name = NULL;
   this->pageid_template.reset(default_pageid_template("p"));
+  this->n_jobs = 1;
 }
 
 namespace string
@@ -184,10 +185,10 @@ namespace string
 
   template <typename tp>
   tp as(const char *);
-  
+
   template <>
   long as<long>(const char *);
-  
+
   template <typename tp>
   tp as(const std::string &);
 }
@@ -236,7 +237,7 @@ static void parse_hyperlinks_options(std::string s, Config::Hyperlinks &options)
       options.extract = false;
       continue;
     }
-    else if 
+    else if
     (
       it->length() == 7 &&
       (*it)[0] == '#' &&
@@ -388,6 +389,7 @@ void Config::read_config(int argc, char * const argv[])
     OPT_DPI = 'd',
     OPT_HELP = 'h',
     OPT_INDIRECT = 'i',
+    OPT_JOBS = 'j',
     OPT_OUTPUT = 'o',
     OPT_PAGES = 'p',
     OPT_QUIET = 'q',
@@ -411,11 +413,12 @@ void Config::read_config(int argc, char * const argv[])
     OPT_PAGEID_TEMPLATE,
     OPT_PAGE_SIZE,
     OPT_PAGE_TITLE_TEMPLATE,
-    OPT_TEXT_LINES,
     OPT_TEXT_CROP,
+    OPT_TEXT_FILTER,
+    OPT_TEXT_LINES,
     OPT_TEXT_NONE,
-    OPT_TEXT_WORDS,
     OPT_TEXT_NO_NFKC,
+    OPT_TEXT_WORDS,
     OPT_VERBATIM_METADATA,
     OPT_VERSION,
   };
@@ -428,10 +431,12 @@ void Config::read_config(int argc, char * const argv[])
     { "crop-text", 0, 0, OPT_TEXT_CROP },
     { "dpi", 1, 0, OPT_DPI },
     { "fg-colors", 1, 0, OPT_FG_COLORS },
+    { "filter-text", 1, 0, OPT_TEXT_FILTER },
     { "guess-dpi", 0, 0, OPT_GUESS_DPI },
     { "help", 0, 0, OPT_HELP },
     { "hyperlinks", 1, 0, OPT_HYPERLINKS },
     { "indirect", 0, 0, OPT_INDIRECT },
+    { "jobs", 1, 0, OPT_JOBS },
     { "lines", 0, 0, OPT_TEXT_LINES },
     { "loss-level", 1, 0, OPT_LOSS_ANY },
     { "losslevel", 1, 0, OPT_LOSS_ANY },
@@ -440,10 +445,10 @@ void Config::read_config(int argc, char * const argv[])
     { "monochrome", 0, 0, OPT_MONOCHROME },
     { "no-hyperlinks", 0, 0, OPT_NO_HLINKS },
     { "no-metadata", 0, 0, OPT_NO_METADATA },
+    { "no-nfkc", 0, 0, OPT_TEXT_NO_NFKC },
     { "no-outline", 0, 0, OPT_NO_OUTLINE },
     { "no-render", 0, 0, OPT_NO_RENDER },
     { "no-text", 0, 0, OPT_TEXT_NONE },
-    { "no-nfkc", 0, 0, OPT_TEXT_NO_NFKC },
     { "output", 1, 0, OPT_OUTPUT },
     { "page-size", 1, 0, OPT_PAGE_SIZE },
     { "page-title-template", 1, 0, OPT_PAGE_TITLE_TEMPLATE },
@@ -461,7 +466,7 @@ void Config::read_config(int argc, char * const argv[])
   while (true)
   {
     optindex = 0;
-    c = getopt_long(argc, argv, "i:o:d:qvp:h", options, &optindex);
+    c = getopt_long(argc, argv, "i:o:d:qvp:j:h", options, &optindex);
     if (c < 0)
       break;
     if (c == 0)
@@ -547,6 +552,10 @@ void Config::read_config(int argc, char * const argv[])
     case OPT_TEXT_NO_NFKC:
       this->text_nfkc = false;
       break;
+    case OPT_TEXT_FILTER:
+      this->text_nfkc = false; /* filter normally does some normalization on its own */
+      this->text_filter_command_line = optarg;
+      break;
     case OPT_TEXT_CROP:
       this->text_crop = true;
       break;
@@ -602,6 +611,9 @@ void Config::read_config(int argc, char * const argv[])
         throw PageTitleTemplateParseError();
       }
       break;
+    case OPT_JOBS:
+      this->n_jobs = string::as<int>(optarg);
+      break;
     case OPT_HELP:
       throw NeedHelp();
     case OPT_VERSION:
@@ -647,7 +659,7 @@ void Config::usage(const Config::Error &error) const
     << std::endl <<   "     --bg-subsample=N"
     << std::endl <<   "     --fg-colors=default"
     << std::endl <<   "     --fg-colors=web"
-#ifdef HAVE_GRAPHICSMAGICK
+#if HAVE_GRAPHICSMAGICK
     << std::endl <<   "     --fg-colors=N"
 #endif
     << std::endl <<   "     --monochrome"
@@ -665,8 +677,12 @@ void Config::usage(const Config::Error &error) const
     << std::endl <<   "     --lines"
     << std::endl <<   "     --crop-text"
     << std::endl <<   "     --no-nfkc"
+    << std::endl << _("     --filter-text=COMMAND-LINE")
     << std::endl <<   " -p, --pages=..."
     << std::endl <<   " -v, --verbose"
+#if _OPENMP
+    << std::endl <<   " -j, --jobs=N"
+#endif
     << std::endl <<   " -q, --quiet"
     << std::endl <<   " -h, --help"
     << std::endl <<   "     --version"
