@@ -1,4 +1,4 @@
-/* Copyright © 2007, 2008, 2009 Jakub Wilk
+/* Copyright © 2007, 2008, 2009, 2010 Jakub Wilk
  * Copyright © 2009 Mateusz Turcza
  *
  * This package is free software; you can redistribute it and/or modify
@@ -7,8 +7,8 @@
  */
 
 #include <algorithm>
+#include <cassert>
 #include <cmath>
-#include <ctime>
 #include <iomanip>
 #include <iostream>
 #include <limits>
@@ -17,6 +17,9 @@
 #include <set>
 #include <sstream>
 #include <vector>
+#if _OPENMP
+#include <omp.h>
+#endif
 
 #include "pdf-backend.hh"
 #include "pdf-dpi.hh"
@@ -29,6 +32,7 @@
 #include "system.hh"
 #include "version.hh"
 #include "i18n.hh"
+#include "xmp.hh"
 
 Config config;
 
@@ -61,7 +65,7 @@ static int get_page_for_LinkGoTo(pdf::link::GoTo *goto_link, pdf::Catalog *catal
       Ref pageref = dest->getPageRef();
       page = catalog->findPage(pageref.num, pageref.gen);
     }
-    else 
+    else
       page = dest->getPageNum();
     return page;
   }
@@ -121,7 +125,7 @@ protected:
 public:
   explicit Component(File &file, const std::string *title = NULL)
   : file(&file)
-  { 
+  {
     if (title != NULL)
       this->title.reset(new std::string(*title));
     file.close();
@@ -129,7 +133,7 @@ public:
 
   Component(const Component &component)
   : file(component.file)
-  { 
+  {
     if (component.title.get() != NULL)
       this->title.reset(new std::string(*component.title));
   }
@@ -161,13 +165,13 @@ Command &operator <<(Command &command, const Component &component)
 {
   command << *component.file;
   return command;
-} 
+}
 
 std::ostream &operator <<(std::ostream &stream, const Component &component)
 {
   stream << *component.file;
   return stream;
-} 
+}
 
 class ComponentList
 {
@@ -222,7 +226,7 @@ public:
   {
     std::vector<Component*>::reference tmpfile_ptr = this->components.at(n - 1);
     if (tmpfile_ptr == NULL)
-    { 
+    {
       this->files[n - 1] = this->create_file(this->get_file_name(n));
       tmpfile_ptr = new Component(*this->files[n - 1], this->get_title(n));
     }
@@ -256,10 +260,10 @@ protected:
       return;
     *(this->text_comments)
       << std::dec << std::noshowpos
-      << "# T " 
-      << ox << ":" << oy << " " 
+      << "\x01 \x02 " /* special characters will be replaced later */
+      << ox << ":" << oy << " "
       << dx << ":" << dy << " "
-      <<  w << "x" <<  h << std::showpos << x << y << " "
+      <<  w << "\x03" <<  h << std::showpos << x << y << " "
       << "("
       << std::oct;
     for (; len > 0; len--, unistr++)
@@ -271,8 +275,8 @@ protected:
     }
     *(this->text_comments) << ")" << std::endl;
   }
-public:
 
+public:
   pdf::Bool needNonText()
   {
     return !config.no_render;
@@ -280,7 +284,7 @@ public:
 
 #if POPPLER_VERSION < 1101
 
-  void drawImageMask(pdf::gfx::State *state, pdf::Object *object, pdf::Stream *stream, int width, int height, 
+  void drawImageMask(pdf::gfx::State *state, pdf::Object *object, pdf::Stream *stream, int width, int height,
     pdf::Bool invert, pdf::Bool inline_image)
   {
     return;
@@ -305,7 +309,7 @@ public:
 
   void drawSoftMaskedImage(pdf::gfx::State *state, pdf::Object *object, pdf::Stream *stream,
     int width, int height, pdf::gfx::ImageColorMap *color_map, pdf::Stream *mask_stream,
-    int mask_width, int mask_height,	pdf::gfx::ImageColorMap *mask_color_map)
+    int mask_width, int mask_height, pdf::gfx::ImageColorMap *mask_color_map)
   {
     if (is_foreground_color_map(color_map) || config.no_render)
       return;
@@ -379,8 +383,8 @@ public:
     }
     else
     {
-      // Ideally, this should not happen. Some heuristics is required to
-      // determine character width/height.
+      /* Ideally, this should never happen. Some heuristics is required to
+       * determine character width/height: */
       pw = pdx; ph = pdy;
       double font_size = state->getTransformedFontSize();
       if (pw * 4.0 < font_size)
@@ -487,7 +491,7 @@ public:
       expr = sexpr::cons(item, expr);
     }
     else
-    { 
+    {
       sexpr::Ref bexpr = sexpr::nil;
       if (border_color.empty())
         bexpr = sexpr::cons(symbol_xor, bexpr);
@@ -522,7 +526,7 @@ public:
   { }
 
   void fill(pdf::gfx::State *state)
-  { 
+  {
     if (config.no_render)
       return;
     pdf::splash::Path path;
@@ -537,9 +541,9 @@ public:
     this->fill(state);
   }
 
-  MutedRenderer(pdf::splash::Color &paper_color, bool monochrome, const ComponentList &page_files) 
+  MutedRenderer(pdf::splash::Color &paper_color, bool monochrome, const ComponentList &page_files)
   : Renderer(paper_color, monochrome), page_files(page_files)
-  { 
+  {
     this->clear_texts();
   }
 
@@ -555,7 +559,20 @@ public:
 
   const std::string get_texts() const
   {
-    return this->text_comments->str();
+    std::string texts = this->text_comments->str();
+    if (config.text_filter_command_line.length() > 0)
+      texts = Command::filter(config.text_filter_command_line, texts);
+    for (std::string::iterator it = texts.begin(); it != texts.end(); it++)
+      switch (*it)
+      {
+      case '\x01':
+        *it = '#'; break;
+      case '\x02':
+        *it = 'T'; break;
+      case '\x03':
+        *it = 'x'; break;
+      }
+    return texts;
   }
 
   void clear_texts()
@@ -607,7 +624,7 @@ static sexpr::Expr pdf_outline_to_djvu_outline(pdf::Object *node, pdf::Catalog *
         if (!pdf::dict_lookup(current, "Title", &title)->isString())
           throw NoTitleForBookmark();
         title_str = pdf::string_as_utf8(title);
-      } 
+      }
 
       int page;
       {
@@ -671,151 +688,31 @@ static bool pdf_outline_to_djvu_outline(pdf::Document &doc, std::ostream &stream
   return true;
 }
 
-class InvalidDateFormat : public std::runtime_error
-{ 
-public:
-  InvalidDateFormat()
-  : std::runtime_error(_("Invalid date format"))
-  { }
-};
-
-static int scan_date_digits(char * &input, int n)
+static void add_meta_string(const char *key, const std::string &value, std::ostream &stream)
 {
-  int value = 0;
-  for (int i = 0; i < n; i++)
-  {
-    if (*input >= '0' && *input <= '9')
-    {
-      value *= 10;
-      value += *input - '0';
-    }
-    else
-      throw InvalidDateFormat();
-    input++;
-  }
-  return value;
+  sexpr::Ref expr = sexpr::string(value);
+  stream << key << "\t" << expr << std::endl;
 }
 
-static void pdf_metadata_to_djvu_metadata(pdf::Document &doc, std::ostream &stream)
+static void add_meta_date(const char *key, const pdf::Timestamp &value, std::ostream &stream)
 {
-  static const char* string_keys[] = { "Title", "Subject", "Keywords", "Author", "Creator", NULL };
-  static const char* date_keys[] = { "CreationDate", "ModDate", NULL };
-  pdf::OwnedObject info;
-  doc.getDocInfo(&info);
-  if (!info.isDict())
-    return;
-  pdf::Dict *info_dict = info.getDict();
-  for (const char** pkey = string_keys; *pkey; pkey++)
-  {
-    pdf::OwnedObject object;
-    if (!pdf::dict_lookup(info_dict, *pkey, &object)->isString())
-      continue;
-    std::string value = pdf::string_as_utf8(object);
-    sexpr::Ref esc_value = sexpr::string(value);
-    stream << *pkey << "\t" << esc_value << std::endl;
-  }
-  {
-    bool have_producer = false;
-    std::string value;
-    pdf::OwnedObject object;
-    if (pdf::dict_lookup(info_dict, "Producer", &object)->isString())
-    {
-      have_producer = true;
-      value = pdf::string_as_utf8(object);
-      if (config.adjust_metadata && value.length() > 0)
-        value += "\n" PACKAGE_STRING;
-    }
-    else if (config.adjust_metadata)
-    {
-      have_producer = true;
-      value = PACKAGE_STRING;
-    }
-    if (have_producer)
-    {
-      sexpr::Ref esc_value = sexpr::string(value);
-      stream << "Producer\t" << esc_value << std::endl;
-    }
-  }
-  for (const char** pkey = date_keys; *pkey; pkey++)
   try
   {
-    pdf::OwnedObject object;
-    struct tm tms;
-    char tzs; int tzh = 0, tzm = 0;
-    char buffer[32];
-    if (!pdf::dict_lookup(info_dict, *pkey, &object)->isString())
-      continue;
-    char *input = object.getString()->getCString();
-    if (input[0] == 'D' && input[1] == ':')
-      input += 2;
-    tms.tm_isdst = -1;
-    tms.tm_year = scan_date_digits(input, 4) - 1900;
-    tms.tm_mon = (*input ? scan_date_digits(input, 2) : 1) - 1;
-    tms.tm_mday = *input ? scan_date_digits(input, 2) : 1;
-    tms.tm_hour = *input ? scan_date_digits(input, 2) : 0;
-    tms.tm_min  = *input ? scan_date_digits(input, 2) : 0;
-    tms.tm_sec  = *input ? scan_date_digits(input, 2) : 0;
-    switch (*input)
-    {
-    case '\0':
-      tzs = 0;
-      break;
-    case '-':
-    case '+':
-      tzs = *input;
-      input++;
-      tzh = scan_date_digits(input, 2);
-      if (tzh > 23)
-        throw InvalidDateFormat();
-      if (*input == '\'')
-        input++;
-      else
-        throw InvalidDateFormat();
-      tzm = scan_date_digits(input, 2);
-      if (tzm > 59)
-        throw InvalidDateFormat();
-      if (*input == '\'')
-        input++;
-      else
-        throw InvalidDateFormat();
-      break;
-    case 'Z':
-      input++;
-      tzs = '+';
-      tzh = tzm = 0;
-      break;
-    default:
-      throw InvalidDateFormat();
-    }
-    if (*input)
-      throw InvalidDateFormat();
-    if (mktime(&tms) == static_cast<time_t>(-1))
-      throw InvalidDateFormat();
-    // RFC 3339 date format, e.g. "2007-10-27 13:19:59+02:00"
-    if (strftime(buffer, sizeof buffer, "%Y-%m-%d %H:%M:%S", &tms) != 19)
-      throw InvalidDateFormat();
-    std::string tzstring;
-    if (tzs)
-    {
-      std::ostringstream stream;
-      stream 
-        << tzs
-        << std::setw(2) << std::setfill('0') << tzh
-        << ":"
-        << std::setw(2) << std::setfill('0') << tzm
-      ;
-      tzstring = stream.str();
-      if (tzstring.length() != 6)
-        throw InvalidDateFormat();
-    } 
-    stream << *pkey << "\t\"" << buffer << tzstring << "\"" << std::endl;
+    sexpr::Ref sexpr = sexpr::string(value.format(' '));
+    stream << key << "\t" << sexpr << std::endl;
   }
-  catch (InvalidDateFormat &ex)
+  catch (pdf::Timestamp::Invalid)
   {
-    debug(1) << string_printf(_("Warning: metadata[%s] is not a valid date"), *pkey) << std::endl;
+    debug(1) << string_printf(_("Warning: metadata[%s] is not a valid date"), key) << std::endl;
   }
 }
 
+static void pdf_metadata_to_djvu_metadata(pdf::Metadata &metadata, std::ostream &stream)
+{
+  sexpr::GCLock gc_lock;
+  metadata.iterate<std::ostream>(add_meta_string, add_meta_date, stream);
+}
+
 class TemporaryComponentList : public ComponentList
 {
 private:
@@ -866,7 +763,7 @@ protected:
   static std::string full_path(const std::string &base_name)
   {
     return DjVuCommand::dir_name + "/" + base_name;
-  } 
+  }
 public:
   static void set_argv0(const char *argv0);
   explicit DjVuCommand(const std::string &base_name)
@@ -943,7 +840,7 @@ protected:
   std::auto_ptr<IndirectDjVm> indirect_djvm;
   std::auto_ptr<TemporaryFile> index_file;
 public:
-  explicit BundledDjVm(File &output_file) 
+  explicit BundledDjVm(File &output_file)
   : output_file(output_file),
     converter("djvmcvt")
   { }
@@ -962,15 +859,15 @@ protected:
   File &index_file;
   std::vector<Component> components;
   class UnexpectedDjvuSedOutput : public std::runtime_error
-  { 
-  public: 
-    UnexpectedDjvuSedOutput() 
-    : std::runtime_error(_("Unexpected output from djvused")) 
+  {
+  public:
+    UnexpectedDjvuSedOutput()
+    : std::runtime_error(_("Unexpected output from djvused"))
     { };
   };
   void do_create(const std::vector<Component> &components, bool shared_ant = false);
 public:
-  explicit IndirectDjVm(File &index_file) 
+  explicit IndirectDjVm(File &index_file)
   : index_file(index_file)
   { }
 
@@ -1025,7 +922,7 @@ public:
       for (size_t i = 0; i < size - 1; i++)
       {
         /* To make the page include the annotations chunk:
-         * - Create an emphemeric indirect multi-page document with just this
+         * - Create an ephemeral indirect multi-page document with just this
          *   single page.
          * - Add the dummy shared annotation.
          */
@@ -1153,19 +1050,44 @@ static int calculate_dpi(const pdf::dpi::Guess &guess)
     return static_cast<int>(dpi);
 }
 
-static int calculate_dpi(double page_width, double page_height)
+static int calculate_dpi(pdf::Document &doc, int n, bool crop)
 {
+  double page_width, page_height;
+  doc.get_page_size(n, crop, page_width, page_height);
+  if (config.guess_dpi)
+  {
+    pdf::dpi::Guesser dpi_guesser(doc);
+    try
+    {
+      pdf::dpi::Guess guess = dpi_guesser[n];
+      std::ostringstream guess_str;
+      guess_str << guess;
+      debug(2)
+        << string_printf(_("guessed resolution: %s dpi"), guess_str.str().c_str())
+        << std::endl;
+      return calculate_dpi(guess);
+    }
+    catch (pdf::dpi::NoGuess)
+    {
+      debug(2) << _("unable to guess resolution") << std::endl;
+    }
+  }
   if (config.preferred_page_size.first)
   {
     double hdpi = config.preferred_page_size.first / page_width;
     double vdpi = config.preferred_page_size.second / page_height;
     double dpi = std::min(hdpi, vdpi) + 0.5;
+    int int_dpi;
     if (dpi < djvu::min_dpi)
-      return djvu::min_dpi;
+      int_dpi = djvu::min_dpi;
     else if (dpi > djvu::max_dpi)
-      return djvu::max_dpi;
+      int_dpi = djvu::max_dpi;
     else
-      return static_cast<int>(dpi);
+      int_dpi = static_cast<int>(dpi);
+      debug(2)
+        << string_printf(_("estimated resolution: %d dpi"), int_dpi)
+        << std::endl;
+    return int_dpi;
   }
   else
     return config.dpi;
@@ -1174,7 +1096,7 @@ static int calculate_dpi(double page_width, double page_height)
 class StdoutIsATerminal : public std::runtime_error
 {
 public:
-  StdoutIsATerminal() 
+  StdoutIsATerminal()
   : std::runtime_error(_("I won't write DjVu data to a terminal."))
   { }
 };
@@ -1230,6 +1152,17 @@ static int xmain(int argc, char * const argv[])
     exit(1);
   }
 
+#if _OPENMP
+  if (config.n_jobs >= 1)
+    omp_set_num_threads(config.n_jobs);
+#else
+  if (config.n_jobs != 1)
+  {
+    debug(1) << string_printf(_("Warning: %s"), _("Multi-threading is not supported.")) << std::endl;
+    config.n_jobs = 1;
+  }
+#endif
+
   if (config.output_stdout && is_stream_a_tty(std::cout))
     throw StdoutIsATerminal();
 
@@ -1237,23 +1170,23 @@ static int xmain(int argc, char * const argv[])
   environment.set_antialias(config.antialias);
 
   size_t pdf_size;
-  { 
+  {
     std::ifstream ifs(config.file_name, std::ifstream::in | std::ifstream::binary);
     ifs.seekg(0, std::ios::end);
     pdf_size = ifs.tellg();
     ifs.close();
   }
 
-  pdf::Document doc(config.file_name);
+  std::auto_ptr<pdf::Document> doc(new pdf::Document(config.file_name));
 
   pdf::splash::Color paper_color;
   pdf::set_color(paper_color, 0xff, 0xff, 0xff);
 
-  size_t n_pixels = 0;
-  size_t djvu_pages_size = 0;
-  int n_pages = doc.getNumPages();
-  unsigned int page_counter = 0;
+  intmax_t n_pixels = 0;
+  intmax_t djvu_pages_size = 0;
+  int n_pages = doc->getNumPages();
   PageMap page_map;
+  std::vector<int> page_numbers;
   std::auto_ptr<const Directory> output_dir;
   std::auto_ptr<File> output_file;
   std::auto_ptr<DjVm> djvm;
@@ -1281,7 +1214,7 @@ static int xmain(int argc, char * const argv[])
     std::string index_file_name = "index.djvu";
     try
     {
-      // For compatibility reasons, check if it's a directory.
+      /* For compatibility reasons, check if it's a directory: */
       output_dir.reset(new Directory(config.output));
     }
     catch (OSError &no_such_directory_exception)
@@ -1303,7 +1236,7 @@ static int xmain(int argc, char * const argv[])
         throw;
       if (config_output_not_a_dir)
       {
-        // Nope, it's not a directory, it must be a file.
+        /* No, it's not a directory, it must be a file: */
         std::string output_directory_name;
         split_path(config.output, output_directory_name, index_file_name);
         if (index_file_name.length() == 0)
@@ -1317,67 +1250,72 @@ static int xmain(int argc, char * const argv[])
   }
   if (config.pages.size() == 0)
     config.pages.push_back(std::make_pair(1, n_pages));
-  int opage = 1;
-  for (
-    std::vector< std::pair<int, int> >::iterator page_range = config.pages.begin();
-    page_range != config.pages.end(); page_range++)
-  for (int ipage = page_range->first; ipage <= n_pages && ipage <= page_range->second; ipage++)
-  {
-    page_map.set(ipage, opage);
-    opage++;
-  }
-
-  std::auto_ptr<pdf::Renderer> out1;
-  std::auto_ptr<MutedRenderer> outm, outs;
-  
-  out1.reset(new pdf::Renderer(paper_color, config.monochrome));
-  outm.reset(new MutedRenderer(paper_color, config.monochrome, *page_files));
-  out1->startDoc(doc.getXRef());
-  outm->startDoc(doc.getXRef());
-  if (!config.monochrome)
-  {
-    outs.reset(new MutedRenderer(paper_color, config.monochrome, *page_files));
-    outs->startDoc(doc.getXRef());
-  }
-  debug(1) << doc.getFileName()->getCString() << ":" << std::endl;
-  bool crop = !config.use_media_box;
-  debug(0)++;
   for (
     std::vector< std::pair<int, int> >::iterator page_range = config.pages.begin();
     page_range != config.pages.end(); page_range++)
   for (int n = page_range->first; n <= n_pages && n <= page_range->second; n++)
   {
-    page_counter++;
+    static int i = 1;
+    page_map.set(i, n);
+    page_numbers.push_back(n);
+    i++;
+  }
+
+  if (page_numbers.size() == 0)
+    throw Config::NoPagesSelected();
+
+  std::auto_ptr<pdf::Renderer> out1;
+  std::auto_ptr<MutedRenderer> outm, outs;
+
+  debug(1) << doc->getFileName()->getCString() << ":" << std::endl;
+  bool crop = !config.use_media_box;
+  debug(0)++;
+  #pragma omp parallel for private(out1, outm, outs, doc) reduction(+: djvu_pages_size)
+  for (size_t i = 0; i < page_numbers.size(); i++)
+  {
+    int n = page_numbers[i];
+    if (out1.get() == NULL)
+    {
+#if _OPENMP
+      assert(doc.get() == NULL);
+#endif
+      doc.reset(new pdf::Document(config.file_name));
+      assert(out1.get() == NULL);
+      out1.reset(new pdf::Renderer(paper_color, config.monochrome));
+      out1->startDoc(doc->getXRef());
+      assert(outm.get() == NULL);
+      outm.reset(new MutedRenderer(paper_color, config.monochrome, *page_files));
+      outm->startDoc(doc->getXRef());
+      assert(outs.get() == NULL);
+      if (!config.monochrome)
+      {
+        outs.reset(new MutedRenderer(paper_color, config.monochrome, *page_files));
+        outs->startDoc(doc->getXRef());
+      }
+    }
+    assert(doc.get() != NULL);
+    assert(out1.get() != NULL);
+    assert(outm.get() != NULL);
+    if (!config.monochrome)
+      assert(outs.get() != NULL);
     Component &component = (*page_files)[n];
-    debug(1) << string_printf(_("page #%d -> #%d"), n, page_map.get(n));
-    debug(2) << ":";
-    debug(1) << std::endl;
+    #pragma omp critical
+    {
+      debug(1) << string_printf(_("page #%d -> #%d"), n, page_map.get(n));
+      if (config.n_jobs != 1)
+        debug(2) << ":";
+      debug(1) << std::endl;
+    }
+#if _OPENMP
+/* Multi-threading would interact badly with logging. Disable it for now. */
+#define debug(x) if (config.n_jobs == 1) (debug)(x)
+#endif
     debug(0)++;
     debug(3) << _("rendering page (1st pass)") << std::endl;
     double page_width, page_height;
-    doc.get_page_size(n, crop, page_width, page_height);
-    int dpi = 0;
-    if (config.guess_dpi)
-    {
-      pdf::dpi::Guesser dpi_guesser(doc);
-      try
-      {
-        pdf::dpi::Guess guess = dpi_guesser[n];
-        std::ostringstream guess_str;
-        guess_str << guess;
-        debug(2)
-          << string_printf(_("guessed resoultion: %s dpi"), guess_str.str().c_str())
-          << std::endl;
-        dpi = calculate_dpi(guess);
-      }
-      catch (pdf::dpi::NoGuess)
-      {
-         debug(2) << _("unable to guess resolution") << std::endl;
-      }
-    }
-    if (dpi == 0)
-      dpi = calculate_dpi(page_width, page_height);
-    doc.display_page(outm.get(), n, dpi, dpi, crop, true);
+    doc->get_page_size(n, crop, page_width, page_height);
+    int dpi = calculate_dpi(*doc, n, crop);
+    doc->display_page(outm.get(), n, dpi, dpi, crop, true);
     int width = outm->getBitmapWidth();
     int height = outm->getBitmapHeight();
     n_pixels += width * height;
@@ -1385,7 +1323,7 @@ static int xmain(int argc, char * const argv[])
     if (!config.no_render)
     {
       debug(3) << _("rendering page (2nd pass)") << std::endl;
-      doc.display_page(out1.get(), n, dpi, dpi, crop, false);
+      doc->display_page(out1.get(), n, dpi, dpi, crop, false);
     }
     debug(3) << _("preparing data for `csepdjvu`") << std::endl;
     debug(0)++;
@@ -1404,7 +1342,7 @@ static int xmain(int argc, char * const argv[])
       double hdpi = sub_width / page_width;
       double vdpi = sub_height / page_height;
       debug(3) << _("rendering background image") << std::endl;
-      doc.display_page(outs.get(), n, hdpi, vdpi, crop, true);
+      doc->display_page(outs.get(), n, hdpi, vdpi, crop, true);
       if (sub_width != outs->getBitmapWidth())
         throw std::logic_error(_("Unexpected subsampled bitmap width"));
       if (sub_height != outs->getBitmapHeight())
@@ -1417,13 +1355,12 @@ static int xmain(int argc, char * const argv[])
       outs->clear_texts();
       outs->clear_annotations();
     }
-    else  
+    else
     {
       nonwhite_background_color = (background_color[0] & background_color[1] & background_color[2] & 0xff) != 0xff;
       if (nonwhite_background_color)
-      {
-        // Dummy background just to assure FGbz chunks.
-        // It will be replaced later.
+      { /* Create a dummy background, just to assure existence of FGbz chunks.
+         * The background chunk will be replaced later: */
         int sub_width, sub_height;
         calculate_subsampled_size(width, height, 12, sub_width, sub_height);
         debug(3) << _("storing dummy background image") << std::endl;
@@ -1454,18 +1391,18 @@ static int xmain(int argc, char * const argv[])
       csepdjvu << sep_file << component;
       csepdjvu();
     }
-    *djvm << component;
     const bool should_have_fgbz = has_background || has_foreground || nonwhite_background_color;
     const bool need_reassemble =
       config.no_render
-      ? true
+      ? false
       : (config.monochrome || nonwhite_background_color || !should_have_fgbz);
     TemporaryFile sed_file;
     if (need_reassemble)
     {
       TemporaryFile sjbz_file, fgbz_file, bg44_file;
       if (!config.monochrome)
-      { 
+      { /* Extract FGbz and BG44 image chunks, to that they can be mangled and
+         * re-assembled later: */
         debug(3) << _("recovering images with `djvuextract`") << std::endl;
         DjVuCommand djvuextract("djvuextract");
         djvuextract << component;
@@ -1477,7 +1414,7 @@ static int xmain(int argc, char * const argv[])
         djvuextract(config.verbose < 3);
       }
       if (config.monochrome)
-      {
+      { /* Use cjb2 for lossy compression: */
         TemporaryFile pbm_file;
         debug(3) << _("encoding monochrome image with `cjb2`") << std::endl;
         DjVuCommand cjb2("cjb2");
@@ -1492,7 +1429,7 @@ static int xmain(int argc, char * const argv[])
       {
         TemporaryFile c44_file;
         c44_file.close();
-        {
+        { /* Create solid-color PPM image with subsample ratio 12: */
           TemporaryFile ppm_file;
           debug(3) << _("creating new background image with `c44`") << std::endl;
           DjVuCommand c44("c44");
@@ -1510,7 +1447,7 @@ static int xmain(int argc, char * const argv[])
           ppm_file.close();
           c44();
         }
-        {
+        { /* Replace previous (dummy) BG44 chunk with the newly created one: */
           debug(3) << _("recovering image chunks with `djvuextract`") << std::endl;
           DjVuCommand djvuextract("djvuextract");
           djvuextract << c44_file << std::string("BG44=") + std::string(bg44_file);
@@ -1518,13 +1455,13 @@ static int xmain(int argc, char * const argv[])
         }
       }
       if (has_text)
-      {
+      { /* Extract hidden text layer (as created by csepdjvu); save it into the sed file: */
         debug(3) << _("recovering text with `djvused`") << std::endl;
         DjVuCommand djvused("djvused");
         djvused << component << "-e" << "output-txt";
         djvused(sed_file);
       }
-      {
+      { /* Re-assemble new DjVu using previously mangled chunks: */
         debug(3) << _("re-assembling page with `djvumake`") << std::endl;
         DjVuCommand djvumake("djvumake");
         std::ostringstream info;
@@ -1540,7 +1477,7 @@ static int xmain(int argc, char * const argv[])
         djvumake();
       }
     }
-    {
+    { /* Extract annotations (hyperlinks); save it into the sed file: */
       debug(3) << _("extracting annotations") << std::endl;
       const std::vector<sexpr::Ref> &annotations = outm->get_annotations();
       sed_file << "select 1" << std::endl << "set-ant" << std::endl;
@@ -1550,7 +1487,7 @@ static int xmain(int argc, char * const argv[])
       outm->clear_annotations();
     }
     sed_file.close();
-    {
+    { /* Add per-page non-raster data into the DjVu file: */
       debug(3) << _("adding non-raster data with `djvused`") << std::endl;
       DjVuCommand djvused("djvused");
       djvused << component << "-s" << "-f" << sed_file;
@@ -1564,16 +1501,35 @@ static int xmain(int argc, char * const argv[])
       djvu_pages_size += page_size;
     }
     debug(0)--;
+#if _OPENMP
+#undef debug
+#endif
+  }
+  /* Separate loop to avoid parallelization: */
+  for (size_t i = 0; i < page_numbers.size(); i++)
+  {
+    int n = page_numbers[i];
+    *djvm << (*page_files)[n];
   }
-  if (page_counter == 0)
-    throw Config::NoPagesSelected();
   djvm->create();
   if (config.extract_metadata)
   {
     TemporaryFile sed_file;
+    pdf::Metadata metadata(*doc);
     debug(3) << _("extracting XMP metadata") << std::endl;
     {
-      const std::string &xmp_bytes = doc.get_xmp();
+      std::string xmp_bytes = doc->get_xmp();
+      debug(0)++;
+      if (config.adjust_metadata)
+        try
+        {
+          xmp_bytes = xmp::transform(xmp_bytes, metadata);
+        }
+        catch (xmp::XmlError &ex)
+        {
+          debug(1) << string_printf(_("Warning: %s"), ex.what()) << std::endl;
+        }
+      debug(0)--;
       if (xmp_bytes.length())
       {
         sexpr::GCLock gc_lock;
@@ -1587,16 +1543,10 @@ static int xmain(int argc, char * const argv[])
           << xmp << std::endl
           << "." << std::endl;
       }
-      else
-      {
-        debug(0)++;
-        debug(3) << _("no XMP metadata") << std::endl;
-        debug(0)--;
-      }
     }
     debug(3) << _("extracting document-information metadata") << std::endl;
     sed_file << "set-meta" << std::endl;
-    pdf_metadata_to_djvu_metadata(doc, sed_file);
+    pdf_metadata_to_djvu_metadata(metadata, sed_file);
     sed_file << "." << std::endl;
     sed_file.close();
     djvm->set_metadata(sed_file);
@@ -1608,12 +1558,12 @@ static int xmain(int argc, char * const argv[])
     debug(3) << _("extracting document outline") << std::endl;
     if (config.format == config.FORMAT_BUNDLED)
     {
-      // Shared annotations chunk in necessary to preserve multi-file document structure.
-      // (Single-file documents cannot contain document outline.)
+      /* Shared annotations chunk in necessary to preserve multi-file document
+       * structure. (Single-file documents cannot contain document outline.) */
       sed_file << "create-shared-ant" << std::endl;
     }
     sed_file << "set-outline" << std::endl;
-    nonempty_outline = pdf_outline_to_djvu_outline(doc, sed_file, *page_files);
+    nonempty_outline = pdf_outline_to_djvu_outline(*doc, sed_file, *page_files);
     sed_file << std::endl << "." << std::endl;
     sed_file.close();
     if (nonempty_outline)
@@ -1632,7 +1582,7 @@ static int xmain(int argc, char * const argv[])
       }
       catch (std::ios_base::failure &ex)
       {
-        // Plausibly no shared annotations
+        /* Let's assume that there are no shared annotations. */
       }
     }
     /* Poppler resets the LC_NUMERIC locale settings:
@@ -1645,7 +1595,7 @@ static int xmain(int argc, char * const argv[])
     double ratio = 1.0 * pdf_size / djvu_size;
     double percent_saved = (1.0 * pdf_size - djvu_size) * 100 / pdf_size;
     debug(0)--;
-    debug(1) 
+    debug(1)
       << string_printf(
            _("%.3f bits/pixel; %.3f:1, %.2f%% saved, %zu bytes in, %zu bytes out"),
            bpp, ratio, percent_saved, pdf_size, djvu_size
